"""
Collections of bootstrap methods and classes.

How to use
----------

test

"""

__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2020-02-07'


import os
from collections import Iterator, Iterable
from itertools import chain
from typing import Union, List

import numpy as np
import xarray as xr

from mlair.data_handler.abstract_data_handler import AbstractDataHandler
from mlair.helpers.helpers import to_list


class BootstrapIterator(Iterator):

    _position: int = None

    def __init__(self, data: "Bootstraps", method, return_reshaped=False):
        assert isinstance(data, Bootstraps)
        self._data = data
        self._dimension = data.bootstrap_dimension
        self.boot_dim = "boots"
        self._method = method
        self._return_reshaped = return_reshaped
        self._collection = self.create_collection(self._data.data, self._dimension)
        self._position = 0

    def __next__(self):
        """Return next element or stop iteration."""
        raise NotImplementedError

    @classmethod
    def create_collection(cls, data, dim):
        raise NotImplementedError

    def _reshape(self, d):
        if self._return_reshaped:
            if isinstance(d, list):
                return list(map(lambda x: self._reshape(x), d))
                # return list(map(lambda x: np.rollaxis(x, -1, 0).reshape(x.shape[0] * x.shape[-1], *x.shape[1:-1]), d))
            else:
                shape = d.shape
                return np.rollaxis(d, -1, 0).reshape(shape[0] * shape[-1], *shape[1:-1])
        else:
            return d

    def _to_numpy(self, d):
        if isinstance(d, list):
            return list(map(lambda x: self._to_numpy(x), d))
        else:
            return d.values

    def apply_bootstrap_method(self, data: np.ndarray) -> Union[np.ndarray, List[np.ndarray]]:
        """
        Apply predefined bootstrap method from given data.

        :param data: data to apply bootstrap method on
        :return: processed data as numpy array
        """
        if isinstance(data, list):
            return list(map(lambda x: self.apply_bootstrap_method(x.values), data))
        else:
            return self._method.apply(data)


class BootstrapIteratorSingleInput(BootstrapIterator):
    _position: int = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __next__(self):
        """Return next element or stop iteration."""
        try:
            index, dimension = self._collection[self._position]
            nboot = self._data.number_of_bootstraps
            _X, _Y = self._data.data.get_data(as_numpy=False)
            _X = list(map(lambda x: x.expand_dims({self.boot_dim: range(nboot)}, axis=-1), _X))
            _Y = _Y.expand_dims({self.boot_dim: range(nboot)}, axis=-1)
            single_variable = _X[index].sel({self._dimension: [dimension]})
            bootstrapped_variable = self.apply_bootstrap_method(single_variable.values)
            bootstrapped_data = xr.DataArray(bootstrapped_variable, coords=single_variable.coords,
                                             dims=single_variable.dims)
            _X[index] = bootstrapped_data.combine_first(_X[index]).reindex_like(_X[index])
            self._position += 1
        except IndexError:
            raise StopIteration()
        _X, _Y = self._to_numpy(_X), self._to_numpy(_Y)
        return self._reshape(_X), self._reshape(_Y), (index, dimension)

    @classmethod
    def create_collection(cls, data, dim):
        l = []
        for i, x in enumerate(data.get_X(as_numpy=False)):
            l.append(list(map(lambda y: (i, y), x.indexes[dim])))
        return list(chain(*l))


class BootstrapIteratorVariable(BootstrapIterator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __next__(self):
        """Return next element or stop iteration."""
        try:
            dimension = self._collection[self._position]
            nboot = self._data.number_of_bootstraps
            _X, _Y = self._data.data.get_data(as_numpy=False)
            _X = list(map(lambda x: x.expand_dims({self.boot_dim: range(nboot)}, axis=-1), _X))
            _Y = _Y.expand_dims({self.boot_dim: range(nboot)}, axis=-1)
            for index in range(len(_X)):
                if dimension in _X[index].coords[self._dimension]:
                    single_variable = _X[index].sel({self._dimension: [dimension]})
                    bootstrapped_variable = self.apply_bootstrap_method(single_variable.values)
                    bootstrapped_data = xr.DataArray(bootstrapped_variable, coords=single_variable.coords,
                                                     dims=single_variable.dims)
                    _X[index] = bootstrapped_data.combine_first(_X[index]).transpose(*_X[index].dims)
            self._position += 1
        except IndexError:
            raise StopIteration()
        _X, _Y = self._to_numpy(_X), self._to_numpy(_Y)
        return self._reshape(_X), self._reshape(_Y), (None, dimension)

    @classmethod
    def create_collection(cls, data, dim):
        l = set()
        for i, x in enumerate(data.get_X(as_numpy=False)):
            l.update(x.indexes[dim].to_list())
        return to_list(l)


class BootstrapIteratorBranch(BootstrapIterator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __next__(self):
        try:
            index = self._collection[self._position]
            nboot = self._data.number_of_bootstraps
            _X, _Y = self._data.data.get_data(as_numpy=False)
            _X = list(map(lambda x: x.expand_dims({self.boot_dim: range(nboot)}, axis=-1), _X))
            _Y = _Y.expand_dims({self.boot_dim: range(nboot)}, axis=-1)
            for dimension in _X[index].coords[self._dimension].values:
                single_variable = _X[index].sel({self._dimension: [dimension]})
                bootstrapped_variable = self.apply_bootstrap_method(single_variable.values)
                bootstrapped_data = xr.DataArray(bootstrapped_variable, coords=single_variable.coords,
                                                 dims=single_variable.dims)
                _X[index] = bootstrapped_data.combine_first(_X[index]).transpose(*_X[index].dims)
            self._position += 1
        except IndexError:
            raise StopIteration()
        _X, _Y = self._to_numpy(_X), self._to_numpy(_Y)
        return self._reshape(_X), self._reshape(_Y), (None, index)

    @classmethod
    def create_collection(cls, data, dim):
        return list(range(len(data.get_X(as_numpy=False))))


class ShuffleBootstraps:

    @staticmethod
    def apply(data):
        size = data.shape
        return np.random.choice(data.reshape(-1, ), size=size)


class MeanBootstraps:

    def __init__(self, mean):
        self._mean = mean

    def apply(self, data):
        return np.ones_like(data) * self._mean


class Bootstraps(Iterable):
    """
    Main class to perform bootstrap operations.

    This class requires a data handler following the definition of the AbstractDataHandler, the number of bootstraps
    to create and the dimension along this bootstrapping is performed (default dimension is `variables`).

    When iterating on this class, it returns the bootstrapped X, Y and a tuple with (position of variable in X, name of
    this variable). The tuple is interesting if X consists on mutliple input streams X_i (e.g. two or more stations)
    because it shows which variable of which input X_i has been bootstrapped. All bootstrap combinations can be
    retrieved by calling the .bootstraps() method. Further more, by calling the .get_orig_prediction() this class
    imitates according to the set number of bootstraps the original prediction.

    As bootstrap method, this class can currently make use of the ShuffleBoostraps class that uses drawing with
    replacement to destroy the variables information by keeping its statistical properties. Use `bootstrap="shuffle"` to
    call this method. Another method is the zero mean bootstrapping triggered by `bootstrap="zero_mean"` and performed
    by the MeanBootstraps class. This method destroy the variable's information by a mode collapse to constant value of
    zero. In case, the variable is normalized with a zero mean, this is equivalent to a mode collapse to the variable's
    mean value. Statistics in general are not conserved in this case, but the mean value of course. A custom mean value
    for bootstrapping is currently not supported.
    """

    def __init__(self, data: AbstractDataHandler, number_of_bootstraps: int = 10,
                 bootstrap_dimension: str = "variables", bootstrap_type="singleinput", bootstrap_method="shuffle"):
        """
        Create iterable class to be ready to iter.

        :param data: a data generator object to get data / history
        :param number_of_bootstraps: the number of bootstrap realisations
        """
        self.data = data
        self.number_of_bootstraps = number_of_bootstraps if bootstrap_method == "shuffle" else 1
        self.bootstrap_dimension = bootstrap_dimension
        self.bootstrap_method = {"shuffle": ShuffleBootstraps(),
                                 "zero_mean": MeanBootstraps(mean=0)}.get(
            bootstrap_method)  # todo adjust number of bootstraps if mean bootstrapping
        self.BootstrapIterator = {"singleinput": BootstrapIteratorSingleInput,
                                  "branch": BootstrapIteratorBranch,
                                  "variable": BootstrapIteratorVariable}.get(bootstrap_type,
                                                                             BootstrapIteratorSingleInput)

    def __iter__(self):
        return self.BootstrapIterator(self, self.bootstrap_method)

    def __len__(self):
        return len(self.BootstrapIterator.create_collection(self.data, self.bootstrap_dimension))

    def bootstraps(self):
        return self.BootstrapIterator.create_collection(self.data, self.bootstrap_dimension)

    def get_orig_prediction(self, path: str, file_name: str, prediction_name: str = "CNN") -> np.ndarray:
        """
        Repeat predictions from given file(_name) in path by the number of boots.

        :param path: path to file
        :param file_name: file name
        :param prediction_name: name of the prediction to select from loaded file (default CNN)
        :return: repeated predictions
        """
        file = os.path.join(path, file_name)
        prediction = xr.open_dataarray(file).sel(type=prediction_name).squeeze()
        vals = np.tile(prediction.data, (self.number_of_bootstraps, 1))
        return vals[~np.isnan(vals).any(axis=1), :]
